#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

class datetime():
    def __init__(self):
        pass
    
    @staticmethod
    def utcnow():
        return datetime()

    def strftime(self, format):
        return "2018-10-14 12:12:12"


#Copyright 2016 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

"""
Sensor handling in Kervi is split in two parts. 
The first part is sensor device drivers that handles physical access to sensor hardware.
The second part is the Sensor class that reads a sensor device and triggers events, store readings to DB.
"""

import time
from kervi.controllers import Controller
from kervi.core.utility.thread import KerviThread
from kervi.spine import Spine
from kervi.values import NumberValue, ColorValue, KerviValue
from kervi.config import Configuration
#from kervi.actions import action
#from kervi.settings import Settings

class Sensor(Controller):
    r"""
    Sensor is a class that exposes a sensor device as a KerviValue.
    The Sensor class polls the associated sensor device and updates it self when
    the value of the sensor device change.

    It is possible to link to other KerviValues and dashboards.

    Some sensor devices are multi dimensional and each dimension are reached by a numeric index of the sensor it self.

    :param sensor_id:
            Id of the sensor. This id is never displayed it is used to reference the sensor in code.
    :type sensor_id: ``str``

    :param name:
            Name of the sensor.
    :type name: ``str``

    :param device:
        The sensor device that should be monitored. Could be one of the sensors from the kervi device library
        or a sensor device driver that inherits from kervi.hal.SensorDeviceDriver
    :type device: ``SensorDeviceDriver``

    :Keyword Arguments:
        * **polling_interval** (``float``) --  Polling interval in seconds. Zero disables polling.        

    """
    def __init__(self, sensor_id, name, device=None, **kwargs):
        Controller.__init__(self, sensor_id, name, **kwargs)
        self._device = device
        self._sub_sensors = []
        self._dimensions = 1
        self._index = kwargs.pop("index", -1)
        self._enabled = None

        if device.value_type == "color":
            self._sensor_value = self.outputs.add("value", name, ColorValue)
        elif device.value_type == "number":
            self._sensor_value = self.outputs.add(sensor_id, name, NumberValue)
        else:
            raise ValueError("Can not handle device value type: " + str(device.value_type))
        
        if self._device:
            self.value_type = self._device.type
            self.value_unit = self._device.unit
            self._sensor_value.type = self._device.type
            self._sensor_value.unit = self._device.unit
            self._sensor_value.min = self._device.min
            self._sensor_value.max = self._device.max
            if self._index == -1:
                self._dimensions = self._device.dimensions
                self._dimension_labels = self._device.dimension_labels
        
        if self._dimensions > 1:
            count = 0
            for label in self._dimension_labels:
                sub_sensor = Sensor(
                        self.component_id + "." + label,
                        label,
                        self._device,
                        use_thread=False,
                        #parent=self,
                        index=count,
                        **kwargs
                    )
                sub_sensor.value_unit = self.value_unit
                self._sub_sensors += [
                    sub_sensor
                ]
                count += 1
        
        if kwargs.get("use_thread", True):
            self._sensor_thread = _SensorThread(self, kwargs.get("polling_interval", 1))
        else:
            self._sensor_thread = None

    def __getitem__(self, sub_sensor):
        if self._dimensions == 1:
            return self
        return self._sub_sensors[sub_sensor]

    
    @property
    def enabled(self):
        return self._enabled == True

    @enabled.setter
    def enabled(self, value):
        self._enabled = value
        if self._dimensions > 1:
            for dimension in range(0, self._dimensions):
                self._sub_sensors[dimension].enable = value

    @property
    def polling_interval(self):
        """
        The polling interval of the sensor in seconds.

        :type: ``float``
        """
        return self._sensor_thread.reading_interval

    @polling_interval.setter
    def reading_interval(self, interval):
        self._sensor_thread.reading_interval = interval

    @property
    def sensor_id(self):
        return self.component_id

    @property
    def device(self):
        """
        The device that is linked to this sensor class.
        It may be a hardware device from the kervi device library
        or a class that inherits from kervi.hal.SensorDeviceDriver
        """
        return self._device

    @device.setter
    def device(self, device):
        self._device = device

    def link_to_dashboard(self, dashboard_id=None, panel_id=None, **kwargs):
        r"""
        Links the sensor to a dashboard.

        :param dashboard_id: Id of the dashboard to link to.
                Enter a * if the sensor should be linked to all dashboards.
        :type dashboard_id: ``str``

        :param panel_id:  Id of the panel to link to.
                This is the id of a panel you have added your self to a dashboard or one of the
                system panels *sys-header*, *header* or *footer*
        :type panel_id: ``str``
        
        
        :Keyword Arguments:
                        
            * **link_to_header** (``str``) -- Link this input to header of the panel.

            * **label_icon** (``str``) -- Icon that should be displayed together with label. All Font Awesome icons are valid just enter the name of the icon without *fa-*

            * **label** (``str``) -- Label text, default value is the name of the sensor.

            * **flat** (``bool``) -- Flat look and feel.

            * **inline** (``bool``) -- Display value, sparkline and label in its actual size otherwise it occupys the entire with of the panel 

            * **type** (``str``) -- One of the following values *radial_gauge*, *vertical_gauge*, *horizontal_gauge*, *chart* or *value*.

            * **show_sparkline** (``bool``) -- Show a sparkline next to the value.

            * **icon** (``bool``) -- Icon to show. All Font Awesome icons are valid just enter the name of the icon without *fa-*.

            * **show_value** (``bool``) -- Show the numeric value and unit.

            * **label** (``str``) -- Label to show default is the name of the sensor.
        """

        if self._dimensions == 1:
            self._sensor_value.link_to_dashboard(dashboard_id, panel_id, **kwargs)
        else:
            for dimension in range(0, self._dimensions):
                self._sub_sensors[dimension].link_to_dashboard(dashboard_id, panel_id, **kwargs)

    # def _get_info(self, **kwargs):
    #     dimensions = []
    #     if self._dimensions > 1:
    #         for dimension in range(0, self._dimensions):
    #             dimensions += [self._sub_sensors[dimension]._get_component_info()]
    #     return {
    #         "type":self._sensor_value._type,
    #         "subSensors": dimensions,
    #         "isInput": False,
    #         "maxValue":self.max,
    #         "minValue":self.min,
    #         "unit":self.value_unit,
    #         "value":self._sensor_value._value,
    #         "value_type": self._device.value_type,
    #         "ranges":self._sensor_value._event_ranges,
    #         "sparkline":self._sensor_value._sparkline
    #     }

    # @action
    # def enable(self, state=True):
    #     self.enabled = state
    
    # @enable.set_interrupt
    # def enable_interrupt(self):
    #     self.enabled = False
    
    def controller_start(self):
        if self._enabled == None:
            self.enabled = True
    
    def _new_sensor_reading(self, sensor_value):
        """
        Call this method to signal a new sensor reading.
        This method handles DB storage and triggers different events.

        :param value:
            New value to be stored in the system.
        """
        if not self._active and not self._enabled:
            return
        
        if self._dimensions > 1:
            for dimension in range(0, self._dimensions):
                value = sensor_value[dimension]
                self._sub_sensors[dimension]._new_sensor_reading(value)
        else:
            self._sensor_value.value = sensor_value
            
    def _read_sensor(self):
        self._new_sensor_reading(self._device.read_value())

    @property
    def log_values(self):
        """
        Set to true if the values should be logged to DB.
        If false Kervi will hold a small cache in memory
        for the last reading to be used in sparklines and real time charts.
        """
        return self._sensor_value._log_values

    @log_values.setter
    def log_values(self, value):
        self._sensor_value._log_values = value

    # @property
    # def persist_value(self):
    #     """
    #     Set to true if the current value should be saved to db when the Kervi application or module exits.
    #     The value will be restored upon application or module start. 
    #     """
    #     return self._sensor_value._persist_value

    # @persist_value.setter
    # def persist_value(self, do_persist):
    #     self._sensor_value.persist_value = do_persist

    def add_value_event(self, event_value, func, event_type=None, parameters=None, **kwargs):
        """
        Add a function that is called when the value reach or pass the event_value.

        :param event_value:
            A single value or range specified as a tuple.

            If it is a range the function specified in func is called when the value enters the range.

        :type event_value: ``float``, ``string``, ``boolean`` or a tuple of these types.

        :param func:
            Function or lambda expression to be called.
            This function will receive the dynamcic value as a parameter.

        :param event_type:
            String with the value "warning" of "error" or None (default).
            If warning or error is specified the value or range are shown in UI.

        :type event_type: ``str``

        """
        self._sensor_value.add_value_event(event_value, func, event_type, parameters, **kwargs)
    
    def add_normal_range(self, value, message=None, func=None, **kwargs):
        self._sensor_value.add_value_event(value, self._sensor_value._handle_range_event, parameters=[message, func, 3], **kwargs)

    def add_warning_range(self, value, message=None, func=None, **kwargs):
        self._sensor_value.add_value_event(value, self._sensor_value._handle_range_event, event_type="warning", parameters=[message, func, 2], **kwargs)

    def add_error_range(self, value, message=None, func=None, **kwargs):
        self._sensor_value.add_value_event(value, self._sensor_value._handle_range_event, event_type="error", parameters=[message, func, 1], **kwargs)

    @property
    def delta(self):
        """
        Enter how much a the value should change before it triggers changes events and updates links.
        :type: ``float``
        """
        return self._sensor_value.delta

    @delta.setter
    def delta(self, value):
        self._sensor_value.delta = value

    @property
    def value(self):
        return self._sensor_value.value
    
    @property
    def max(self):
        """
        Maximum value.

        :type: ``float``
        """
        return self._sensor_value.max_value

    @max.setter
    def max(self, value):
        self._sensor_value.max_value = value

    @property
    def min(self):
        """
        Minimum value.

        :type: ``float``
        """
        return self._sensor_value.min_value

    @min.setter
    def min(self, value):
        self._sensor_value.min_value = value

    @property
    def unit(self):
        """
        Metric Unit of value.

        :type: ``str``
        """
        return self._sensor_value.unit

    @unit.setter
    def unit(self, value):
        self._sensor_value.unit = value

    @property
    def display_unit(self):
        """
        Display unit of value.

        :type: ``str``
        """
        return self._sensor_value.display_unit

    @display_unit.setter
    def display_unit(self, value):
        self._sensor_value.display_unit = value
    

    def set_ui_parameter(self, name, value):
        self._sensor_value.set_ui_parameter(name, value)

class _SensorThread(KerviThread):
    r"""
    SensorThread is the base class that  polls sensors.
    Add one or more sensors and set polling interval.

    :param sensors:
            Id of the sensor.
            This id is used in other components to reference this sesnor.
    :type sensors: ``str``

    :param reading_interval:
            Polling interval in seconds between
    :type reading_interval: ``float``

    """

    def __init__(self, sensors, reading_interval=1):
        KerviThread.__init__(self)
        self.spine = Spine()
        if self.spine:
            self.spine.register_command_handler("startThreads", self._start_command)
            self.spine.register_command_handler("stopThreads", self._stop_command)
        self.alive = False
        self.reading_interval = reading_interval
        if hasattr(sensors, "__len__"):
            self.sensors = sensors
        else:
            self.sensors = [sensors]

    def new_sensor_reading(self, value, sensor_idx=0):
        self.sensors[sensor_idx]._new_sensor_reading(value)

    def _step(self):
        for sensor in self.sensors:
            sensor._read_sensor()
        self.sensor_step()
        time.sleep(self.reading_interval)

    def sensor_step(self):
        pass

    def _start_command(self):
        if not self.alive:
            self.alive = True
            KerviThread.start(self)

    def _stop_command(self):
        if self.alive:
            self.alive = False
            self.stop()





#Copyright 2016 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.


"""Hardware abstraction layer"""
import time
from kervi.hal import gpio
from kervi.spine import Spine
from kervi.core.utility.thread import KerviThread
import importlib
import os
#import kervi.utility
_DRIVER = None
GPIO = None

HAL_DRIVER_ID = None

def _load(hw_platform="auto"):
    global GPIO, _DRIVER, HAL_DRIVER_ID

    if not _DRIVER:
        hal_modules = {
            "windows": "kervi.platforms.windows",
            #"linux": "kervi.platforms.linux",
            #"darwin": "kervi.platforms.linux",
            "linux(rpi)": "kervi.platforms.raspberry",
            "generic": "kervi.platforms.generic",
            "upython": "kervi.platforms.upython"
        }
        
        if hw_platform=="auto":
            import platform
            system = platform.system().lower()
            if system == "linux":
                try:
                    with open('/proc/cpuinfo') as f:
                        for line in f:
                            line = line.strip()
                            if line.startswith('Hardware') and line.endswith('BCM2708'):
                                system = "linux(rpi)"
                                break
                            elif line.startswith('Hardware') and line.endswith('BCM2835'):
                                system = "linux(rpi)"
                                break
                            elif line.startswith('Hardware') and line.endswith('BCM2836'):
                                system = "linux(rpi)"
                                break
                            elif line.startswith('Hardware') and line.endswith('BCM2837'):
                                system = "linux(rpi)"
                                break
                except:
                    pass
            if not system in hal_modules.keys():
                system = "generic"
        else:
            system = hw_platform.lower()      
            
        if system == "upython":
            import ukervi.platforms.upython as _DRIVER
        elif system in hal_modules.keys():
            module_name = hal_modules[system] 
            _DRIVER = importlib.import_module(module_name)
            HAL_DRIVER_ID = module_name
            GPIO = get_gpio()
            return module_name
        else:
            raise ValueError("Invalid hw_platform. Valid values are: " + str.join(hal_modules.keys()))

def get_user_inputs():
    return _DRIVER.get_user_inputs()

def detect_devices():
    devices = _DRIVER.detect_devices()
    return devices

def get_gpio(gpio_type=None):
    if gpio_type == None:
        return _DRIVER.get_gpio_driver()

def default_i2c_bus():
    if _DRIVER:
        return _DRIVER.default_i2c_bus()
    return 0

def get_i2c(address, bus=default_i2c_bus()):
    return _DRIVER.get_i2c_driver(address, bus)

def get_one_wire(address):
    return _DRIVER.get_one_wire_driver(address)


def get_camera_driver(source = None):
    return _DRIVER.get_camera_driver(source)

def service_commands(commands, app_name, app_id, script_path):
    return _DRIVER.service_commands(commands, app_name, app_id, script_path)


def device_reboot():
    _DRIVER.reboot()

def device_shutdown():
    _DRIVER.shutdown()

#if not _DRIVER:
#    _load()


class SensorDeviceDriver(object):

    @property
    def dimensions(self):
        return 1

    @property
    def dimension_labels(self):
        return []

    @property
    def device_name(self):
        return None

    @property
    def max(self):
        return None

    @property
    def min(self):
        return None

    # @property
    # def sensor_type(self):
    #     raise NotImplementedError

    @property
    def unit(self):
        raise NotImplementedError

    def read_value(self):
        raise NotImplementedError

    @property
    def logger(self):
        return Spine().log

    @property
    def value_type(self):
        return "number"

class I2CaddressOutOfBoundsError(Exception):
    def __init__(self, device_name, address):
        super(I2CaddressOutOfBoundsError, self).__init__(
            '{0} I2C Address Out of Bounds:{1}'.format(device_name, address)
        )

class DeviceChannelOutOfBoundsError(Exception):
    def __init__(self, device_name, channel):
        super(DeviceChannelOutOfBoundsError, self).__init__(
            '{0} Exception: Channel Out of Bounds, channel={1}'.format(device_name, channel)
        )

# Exception class for a DAC value out of bounds
class DACValueOutOfBoundsError(Exception):
    def __init__(self, device_name, channel, value):
        super(DACValueOutOfBoundsError, self).__init__(
            '{0} Exception: DAC Output Value Out of Bounds, channel={1} value={2}'.format(device_name, channel, value)
        )

class I2CSensorDeviceDriver(SensorDeviceDriver):
    def __init__(self, address, bus):
        if address > 0x77:
            raise I2CaddressOutOfBoundsError(self.device_name, address)
        self.i2c = get_i2c(address, bus)

class I2CGPIODeviceDriver(gpio.IGPIODeviceDriver):
    def __init__(self, address, bus, gpio_id):
        gpio.IGPIODeviceDriver.__init__(self, gpio_id)
        if address > 0x77:
            raise I2CaddressOutOfBoundsError(self.device_name, address)
        self.i2c = get_i2c(address, bus)

    @property
    def device_name(self):
        return None

    @property
    def num_gpio(self):
        return 0

    def _bit2(self, src, bit, val):
        bit = 1 << bit
        return (src | bit) if val else (src & ~bit)


    def _validate_channel(self, channel):
        # Raise an exception if pin is outside the range of allowed values.
        if channel < 0 or channel >= self.num_gpio:
            raise DeviceChannelOutOfBoundsError(self.device_name, channel)

class OneWireSensorDeviceDriver(SensorDeviceDriver):
    def __init__(self, address):
        self.one_wire = get_one_wire(address)

class ChannelPollingThread(KerviThread):
    def __init__(self, channel, device, callback, polling_time=.1):
        KerviThread.__init__(self)
        self._callback = callback
        self._channel = channel
        self._device = device
        self._value = None
        self._polling_time = polling_time
        self.alive = False
        self.spine = Spine()
        if self.spine:
            self.spine.register_command_handler("startThreads", self._start_command)
            self.spine.register_command_handler("stopThreads", self._stop_command)

    def _step(self):
        """Private method do not call it directly or override it."""
        try:
            new_value = self._device.get(self._channel)
            if new_value != self._value:
                self._callback(new_value)
                self._value = new_value
            time.sleep(self._polling_time)
        except:
            self.spine.log.exception("_PollingThread")

    def _start_command(self):
        if not self.alive:
            self.alive = True
            KerviThread.start(self)

    def _stop_command(self):
        if self.alive:
            self.alive = False
            self.stop()




#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

"""
A Kervi controller is a class that acts upon input from users or events or the underlaying os.
"""
#import inspect
#from kervi.spine import Spine
#from kervi.utility.thread import KerviThread
from kervi.core.utility.component import KerviComponent
#from kervi.hal import GPIO
from kervi.values.value_list import ValueList
from kervi.values.kervi_value import KerviValue
from kervi.values import NumberValue
from kervi.actions import Actions
from kervi.actions.action import Action, _ActionInterrupt

class Controller(KerviComponent):
    """
    A Kervi controller is a class that acts upon input from users, hardware events or the underlaying os.

    Examples for controllers are motor control, servo control, output to IO.

    A controller my define one or more dynamic values as input and output.

    """
    def __init__(self, controller_id, name, **kwargs):
        KerviComponent.__init__(self, controller_id, "controller", name, **kwargs)
        self.inputs = ValueList(self, True)
        self.outputs = ValueList(self, False)
        self._active = False
        self.type = "unknown"
        self.parameters = {}
        self._ui_parameters = {
            "size": 0,
            "type": "normal",
            "link_to_header": False,
            "icon": None,
            "flat": False,
            "inline": False
        }
        self.spine.register_event_handler("processTerminating", self._on_terminate)
        self.spine.register_event_handler("appReady", self._on_app_ready)
        self.spine.register_event_handler("moduleStarted", self._on_app_ready)
        self.actions = {}

        method_list = [func for func in dir(self)]
        for method_name in method_list:
            try:
                if hasattr(self, method_name):
                    method = getattr(self, method_name)
                    method_qual_name = getattr(method, "__qualname__", None)
                    if not method_qual_name:
                        method_qual_name = self.__class__.__name__ + "." + method_name
                    
                    if Actions.is_unbound(method_qual_name):
                        action_id, name = Actions.get_unbound(method_qual_name)
                        setattr(self, "kervi_action_"+ method.__name__, method)
                        copy_method = getattr(self, "kervi_action_"+ method.__name__)
                        action = Action(copy_method, self.controller_id + "." + action_id, name)
                        Actions.add(action)
                        self.actions[action_id] = action
                        setattr(self, method.__name__, action)
            except KeyError:
                pass

        for method_name in method_list:
            try:
                if hasattr(self, method_name):
                    method = getattr(self, method_name)
                    method_qual_name = getattr(method, "__qualname__", None)
                    if not method_qual_name:
                        method_qual_name = self.__class__.__name__ + "." + method_name
                    if Actions.is_unbound_interrupt(method_qual_name):
                        action_id = Actions.get_unbound_interrupt(method_qual_name)
                        action = self.actions[action_id]
                        action._interrupt = _ActionInterrupt(method)
                        action.set_ui_parameter("interrupt_enabled", True)
            except KeyError:
                pass

    @property
    def controller_id(self):
        return self.component_id

    def link_to_dashboard(self, dashboard_id=None, panel_id=None, **kwargs):
        r"""
        Links this component to a dashboard section all input are displayed.

        :param dashboard_id:
            id of the dashboard to link to.
        :type section_id: str

        :param panel_id:
            id of the section.
        :type panel_id: str


        :param \**kwargs:
            Use the kwargs below to override default values set in ui_parameters

        :Keyword Arguments:
            * *ui_size* (``int``) -- Size of the component in dashboard unit size.
                In order to make the sections and components align correct a dashboard unit is defined.
                Default the dashboard unit is a square that is 150 x 150 pixels.
                The width of the select box is ui_size * dashboard unit size.

            * *type* (``str``) -- Type of select box, dropdown or list.
            * *link_to_header* (``str``) -- Add this component to header of section.
            * *icon* (``str``) -- Icon that should be displayed together with label.
            * *flat* (``bool``) -- Flat look and feel.

        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            panel_id,
            *kwargs
        )

    def _on_app_ready(self, id):
        self._active = True
        self.controller_start()

    def _on_terminate(self, id):
        
        if self._active:
            self._active = False
            try:
                self.spine.log("exit controller:{0}", self.component_id)
            except:
                pass
            finally:
                self.controller_exit()

    def controller_exit(self):
        """
        Abstract method that is called when the kervi application stops.
        """
        pass

    def controller_start(self):
        """
        Abstract method that is called when the entire application is loaded.
        """
    
    def kervi_value_changed(self, source, value):
        if self._active:
            self.input_changed(source)

    def input_changed(self, input_id):
        """
        Abstract method that is called by when one of the controller inputs change.

        You can implement this method if your controller logic needs to respond
        changes in multiple inputs.

        :param changed_input:
            Input that has changed. You can read the value of the changed input
            via the inputs value property.

        :type changed_input: KerviValue
        """
        pass

    def _get_info(self, **kwargs):
        inputs = []
        for key in self.inputs.keys:
            if self.inputs[key]:
                inputs += [self.inputs[key].get_reference()]

        outputs = []
        for key in self.outputs.keys:
            if self.outputs[key]:
                outputs += [self.outputs[key].get_reference()]

        actions=[]
        for action in self.actions:
            actions += [self.actions[action].get_reference()]

        template = None
        try:
            import os.path
            import sys
            modulepath = os.path.dirname(sys.modules[self.__class__.__module__].__file__)
            class_name = self.__class__.__name__
            cpath = os.path.join(modulepath, class_name + ".tmpl.html")
            if os.path.isfile(cpath):
                template_file = open(cpath, 'r')
                template = template_file.read()
        except:
            pass

        return {
            "inputs":inputs,
            "outputs":outputs,
            "actions":actions,
            "template" : template,
            "type": self.type
        }


    @staticmethod
    def create(controller_id, name):
        """Turn class into a kervi controller"""
        def _decorator(cls):
            class _ControllerClass(cls, Controller):
                def __init__(self):
                    Controller.__init__(self, controller_id, name)
                    for key in cls.__dict__.keys():
                        prop = cls.__dict__[key]
                        if isinstance(prop, KerviValue):
                            if prop.is_input:
                                self.inputs._add_internal(key, prop)
                            else:
                                self.outputs._add_internal(key, prop)
                    cls.__init__(self)
            return _ControllerClass
        return _decorator

    @staticmethod
    def input(input_id, name, value_class=NumberValue):
        """Add input to controller"""
        def _init():
            return value_class(
                name,
                input_id=input_id,
                is_input=True,
                index=-1
            )
        def _decorator(cls):
            setattr(cls, input_id, _init())
            return cls
        return _decorator

    @staticmethod
    def output(output_id, name, value_class=NumberValue):
        """Add output to controller"""
        def _init():
            return value_class(
                name,
                input_id=output_id,
                is_input=False,
                index=-1
            )
        def _decorator(cls):
            setattr(cls, output_id, _init())
            return cls
        return _decorator



#Copyright 2018 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

"""
The display module holds classes for handling displays.
"""
import os
import threading
import time
from kervi.controllers import Controller
from kervi.values import *
from kervi.sensors import Sensor
from kervi.values.value_list import ValueList
from kervi.actions import action 
from PIL import Image, ImageDraw


class _PageTimer(threading.Thread):
    def __init__(self, display):
        threading.Thread.__init__(self, name="pageTimer")
        self.deamon = True
        self._display = display
        self._terminated = False

    def stop(self):
        self._terminated = True
    
    def run(self):
        while not self._terminated:
            self._display.activate_page("next")
            time.sleep(self._display.page_change_speed)

class Display(Controller):
    r"""
    The display class manage a display device.



    :param display_id:
            Id of the display. This id is never displayed it is used to reference the display in code.
    :type display_id: ``str``

    :param name:
            Name of the display.
    :type name: ``str``

    :param device:
        The display device that should be used. Could be one of the displays from the kervi device library
        or a display device driver that inherits from kervi.hal.DisplayDeviceDriver
    :type device: ``DisplayDeviceDriver``

    """
    def __init__(self, display_id, name, device_driver):
        Controller.__init__(self, display_id, name)
        self._device = device_driver
        self.text = self.inputs.add("text", "Text", StringValue)
        self._bitmap = None
        self._display_active = False
        self._line_height = 0
        self._font = None
        self._background_color=(0)
        self._text_color = (255)
        self._pages = {}
        self._active_page = None
        self._page_change_speed = 5
        self._page_timer = None
        self._bitmap_lock = threading.Lock()
    
    def add_page(self, page, default = True):
        r"""
        Add a display page to the display.

        :param page:  Page to be added
        :type display_id: ``DisplayPage``

        :param default: True if this page should be shown upon initialization.
            
        :type name: ``bool``
        """
        self._pages[page.page_id] = page
        page._add_display(self)
        if default or not self._active_page:
            self._active_page = page
    
    @property
    def page_change_speed(self):
        return self._page_change_speed

    @page_change_speed.setter
    def page_change_speed(self, value):
        self._page_change_speed = value
    
    @property
    def display_pages(self):
        r"""
        Returns the display pages added to this display 
        """
        return self._pages

    @property
    def active_page(self):
        r"""
        The current active display page or None if no pages are added.
        """
        return self._active_page

    @action
    def activate_page(self, page_id):
        r"""
        Activates a display page. Content of the active page is shown in the display.

        :param page_id:  Id of page to activate
        :type page_id: ``str``

        """
        if page_id == "next":
            page_keys = list(self._pages.keys())
            key_count = len(page_keys)
            if key_count > 0:
                idx = page_keys.index(self._active_page.page_id)
                if idx >= key_count-1:
                    idx = 0
                else:
                    idx += 1
                self._active_page = self._pages[page_keys[idx]]
                self._active_page._render()
                self._page_changed(self._active_page)

        elif page_id in self._pages:
            self._active_page = self._pages[page_id]
            self._page_changed(self._active_page)
        else:
            raise ValueError("Page with {page_id} not found in page list".format(page_id = page_id))

    @property
    def bitmap(self):
        return self._bitmap

    @bitmap.setter
    def bitmap(self, value):
        self._bitmap = value

    @property
    def line_height(self):
        return self._line_height

    @line_height.setter
    def line_height(self, value):
        self._line_height = value

    @property
    def font(self):
        if not self._font:
            self._font = self._get_font()
        return self._font
    
    @action
    def display_active(self, state=True):
        self._display_active = state
        self._device.enable_display(self._display_active)

    @display_active.set_interrupt
    def display_active_interupt(self):
        self._display_active = False
        self._device.enable_display(self._display_active)

    @action
    def activate_page_scroll(self):
        if not self._page_timer:
            self._page_timer = _PageTimer(self)
            self._page_timer.start()

    @activate_page_scroll.set_interrupt
    def activate_page_scroll_interrupt(self):
        if self._page_timer:
            self._page_timer.stop()
            self._page_timer.join()
            self._page_timer = None

    @action
    def scroll_v(self):
        pass

    @action
    def scroll_h(self):
        pass

    def _get_font(self, size=8, name="PressStart2P.ttf"):
        """
        Returns a font that can be used by pil image functions.
        This default font is "SourceSansVariable-Roman" that is available on all platforms.
        """
        import kervi.vision as vision
        from PIL import ImageFont
        vision_path = os.path.dirname(vision.__file__)
        fontpath = os.path.join(vision_path, "fonts", name)
        font = ImageFont.truetype(fontpath, size)
        return font
    
    def _page_changed(self, changed_page):
        if self._active_page == changed_page:
            self.text.value = changed_page.text
    
    def input_changed(self, changed_input):
        if changed_input == self.text:
            if self._device.display_type == "char":
                self._device.message(self.text.value)
            else:
                line_count = len(self.text.value.split("\n"))
                line_height = self._line_height
                if line_height == 0:
                    line_height = int(self._device.height / line_count )
                    if line_height < 16:
                        line_height = 8
                if line_height == 8:
                    font = self._get_font(line_height, "PixelOperatorMono8-Bold.ttf")
                else:
                    font = self._get_font(line_height, "PixelOperatorMono-Bold.ttf")
                image = Image.new('1', size=(self._device.width, self._device.height), color=self._background_color)
                draw = ImageDraw.Draw(image)
                draw.multiline_text((0, 0), self.text.value, font=font, fill=self._text_color, spacing=2)
                self._bitmap_lock.acquire()
                try:
                    self._device.image(image)
                    self._device.display()
                finally:
                    self._bitmap_lock.release()
                
    def controller_start(self):
        if self._active_page:
            self._active_page._render()

    def controller_exit(self):
        if self._page_timer:
            self._page_timer.stop()


class _DisplayLink:
    def __init__(self, transform, value_id, link_id, x=0, y=0, size=0):
        self.transform = transform
        self.link_id = link_id
        self.value_id = value_id
        self.x= x
        self.y = y
        self.size = size
        self.value = ""
        self.unit = ""

    def _set_value(self, value, unit):
        if self.transform:
            self.value = self.transform(value)
        else:
            self.value = value

        self.unit = unit

class DisplayPage(Controller):
    def __init__(self, page_id, name = None):
        Controller.__init__(self, page_id, name)
        self.spine.register_event_handler("valueChanged", self._link_changed_event)
        self._template = ""
        self._links = {}
        self._displays = []
        self._text = None
        self._lines = 1

    def _link_changed_event(self, source_id, value, x):
        #print("lc", source_id, value)
        if source_id in self._links.keys():
            link = self._links[source_id]
            link._set_value(value["display_value"], value["display_unit"])
            self._render()

    def _add_display(self, display):
        self._displays.append(display)
    
    @property
    def line_count(self):
        return self._lines

    @property
    def text(self):
        return self._text
   
    @property
    def page_id(self):
        return self.component_id 

    @property
    def template(self):
        return self._template

    @template.setter
    def template(self, value):
        self._template = value
        self._lines = len(self._template.split("\n"))
    
    def _render(self):
        kwargs = {}
        for link_id in self._links:
            link = self._links[link_id]
            kwargs[link.link_id] = link.value
            kwargs[link.link_id + "_unit"] = link.unit
        self._text = self.template.format(self._template, **kwargs)
        for display in self._displays:
            display._page_changed(self) 
    
    def _transform(self, v, value_format):
        if value_format:
            return str.format("{:" + value_format + "}", v)
        else:
            return str(v)
    
    def input_changed(self, value):
        self._render()


    def link_value(self, source, format=None):
        id = None
        value_id = None
        if isinstance(source, Sensor):
            id = source.sensor_id
            value_id = source._sensor_value.value_id
        elif isinstance(source, KerviValue):
            id = source.value_id
            value_id = id
        elif isinstance(source, str):
            id = source
            value_id = id
        if id:
            self.spine.register_event_handler("valueChanged", self._link_changed_event, value_id)
            self._links[value_id] = _DisplayLink(lambda x: self._transform(x, format), value_id, id)
        else:
            raise ValueError("Source must be a KerviValue, Sensor or string")


#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

import time
from kervi.controllers import Controller
from kervi.values import NumberValue, BooleanValue
from kervi.actions import action
from kervi.controllers import Controller
class MotorSteering(Controller):
    """
    Control the speed and direction of two motors.
    """
    def __init__(self, controller_id="steering", name="Steering"):
        Controller.__init__(self, controller_id, name)
        self.type = "steering"
        self.left_speed = self.outputs.add("left_speed", "Left speed", NumberValue)
        self.right_speed = self.outputs.add("right_speed", "Right speed", NumberValue)

        self.inputs.add("left_encoder", "Left encoder", NumberValue)
        self.inputs.add("right_encoder", "Right encoder", NumberValue)

        self._adjust = 0

        self.speed = self.inputs.add("speed", "Speed", NumberValue)
        self.adaptive_speed = self.inputs.add("adaptive_speed", "Adaptive speed", NumberValue)
        self.direction = self.inputs.add("direction", "Direction", NumberValue)
        self.adaptive_direction = self.inputs.add("Adaptive direction", "Adaptive direction", NumberValue)

    @property
    def adjust(self):
        return self._adjust

    @adjust.setter
    def adjust(self, value):
        self._adjust = value

    @action
    def move(self, speed, **kwargs):
        direction= kwargs.pop("direction",None)
        duration=kwargs.pop("duration",None),
        wheel_rotations=kwargs.pop("wheel_rotations",None)

        if direction:
            self.direction.value = direction

        self.speed.value = speed

    @action
    def stop(self, **kwargs):
        coast = kwargs.pop("break", True)
        self.outputs["left_speed"].value = 0
        self.outputs["right_speed"].value = 0

    @action
    def rotate(self, **kwargs):
        speed = kwargs.pop("speed",None)
        wheel_rotations = kwargs.pop("wheel_rotations",None)
        duration = kwargs.pop("duration",None)
        
        self.spine.log.verbose("steering rotate: %s", speed)
        new_direction = self._adjust
        left_speed = speed * (-new_direction / 100)
        right_speed = speed * (new_direction / 100)

        self.outputs["left_speed"].value = left_speed
        self.outputs["right_speed"].value = right_speed
        if duration:
            time.sleep(duration)

    def _update(self):
        self.spine.log.verbose("steering update: %s %s", self.speed.value, self.direction.value)
        new_direction = self.direction.value + self.adaptive_direction.value + self._adjust
        speed = self.speed.value + self.adaptive_speed.value
        self.spine.log.verbose("steering adaptive update: speed= %s direction=%s", speed, new_direction)
        if new_direction > 0:
            left_speed = speed * (1 - new_direction / 100)
            right_speed = speed
        elif new_direction < 0:
            left_speed = speed
            right_speed = speed * (1 + new_direction / 100)
        elif new_direction == 0:
            left_speed = speed
            right_speed = speed

        self.outputs["left_speed"].value = left_speed
        self.outputs["right_speed"].value = right_speed

    def input_changed(self, changed_input):
        if changed_input in [self.speed, self.direction, self.adaptive_speed]:
            self._update()



#Copyright 2016 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

"""
Module that holds log functionality.
In general you dont need to include this in your kervi code as the spine class holds a reference
to a fully configured log.
"""

import logging
import logging.handlers
import os

VERBOSE = 15

logging.addLevelName(VERBOSE, "VERBOSE")

class BraceMessage(object):
    def __init__(self, fmt, *args):
        self.fmt = fmt
        self.args = args

    def __str__(self):
        try:
            if '{' in self.fmt:
                return self.fmt.format(*self.args)
            else:
                return self.fmt % self.args
        except:
            return self.fmt

class KerviLog(object):
    def __init__(self, name):
        
        self.logger = logging.getLogger(name)
        

    def info(self, message, *args):
        self.logger.info(BraceMessage(message, *args))

    def verbose(self, message, *args, **kwargs):
        if args:
            self.logger._log(VERBOSE, str(BraceMessage(message, *args)), None)
        else:
            self.logger._log(VERBOSE, message, *args)
    
    def warn(self, message, *args):
        self.logger.warning(BraceMessage(message, *args))

    def debug(self, message, *args):
        if args:
            self.logger.debug(BraceMessage(message, *args))
        else:
            self.logger.debug(message)

    def error(self, message, *args):
        self.logger.error(BraceMessage(message, *args))

    def exception(self, message, *args):
        self.logger.exception(BraceMessage(message, *args))

    def fatal(self, message, *args):
        self.logger.fatal(BraceMessage(message, *args))

def init_process_logging(process_name, config, log_queue=None):
    logger = logging.getLogger()
    
    if config.level == "verbose":
        logger.setLevel(VERBOSE)
    elif config.level == "info":
        logger.setLevel(logging.INFO)
    elif config.level == "warning":
        logger.setLevel(logging.WARNING)
    elif config.level == "debug":
        logger.setLevel(logging.DEBUG)    
        
    if log_queue:
        queue_handler = logging.handlers.QueueHandler(log_queue)
        logger.addHandler(queue_handler)
    else:
        if config.resetLog:
            try:
                os.remove(config.file)
            except:
                pass

        file_handler = logging.FileHandler(config.file)
        file_handler.setLevel(logging.DEBUG)
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.ERROR)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        stream_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        logger.addHandler(stream_handler)



#Copyright 2016 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

"""
When Kervi starts it scans for installed Kervi platform drivers and loads the GPIO driver.
In your application you access GPIO via:

.. code:: python

    import kervi.hal.GPIO as GPIO

    #define channel 23 as input
    GPIO.define_as_input(23)
"""
from kervi.values import NumberValue, BooleanValue
from kervi.values.value_list import ValueList
from kervi.core.utility.kervi_logging import KerviLog

CHANNEL_TYPE_ANALOG_IN = 1
CHANNEL_TYPE_ANALOG_OUT = 2
CHANNEL_TYPE_GPIO = 3

class LogicIOChannel(BooleanValue):
    def __init__(self, gpio_device, channel):
        BooleanValue.__init__(self, gpio_device.device_name +" " + str(channel), input_id=str(channel))
        self._device = gpio_device
        self._channel = channel

        self.pwm = ValueList(self, True)
        self.pwm.add("duty_cycle", "Duty cycle", NumberValue)
        self.pwm.add("frequency", "Frequency", NumberValue)
        self.pwm.add("active", "Active", BooleanValue)

        self.pwm["duty_cycle"].min = 0
        self.pwm["duty_cycle"].max = 100

        self.pwm["frequency"].min = 0
        self.pwm["frequency"].max = 100000


    def get(self):
        return self._device.get(self, self._channel)

    def set(self, value):
        if self.is_input:
            self._device.set(self._channel, value)

    def _input_changed(self,v):
        self.value = self._device.get(self._channel)

    def define_as_input(self, pullup=False, bounce_time=200):
        self.is_input = False
        self._device.define_as_input(self._channel, pullup)
        self._device.listen(self._channel, self._input_changed, bounce_time)

    def define_as_output(self):
        self.is_input = True
        self._device.define_as_output(self._channel)

    def define_as_pwm(self, frequency=None, duty_cycle=None):
        self._device.define_as_pwm(self._channel, frequency, duty_cycle)
        if frequency:
            self.pwm["frequency"].value = frequency

        if duty_cycle:
            self.pwm["duty_cycle"].value = duty_cycle
        

    def listen(self, callback, bounce_time=.2):
        self._device.listen(self._channel, callback, bounce_time)

    def listen_rising(self, callback):
        self._device.listen_rising(self._channel, callback)

    def listen_falling(self, callback):
        self._device.listen_faling(self._channel, callback)

    def pwm_start(self, duty_cycle=None, frequency=None):
        self._device.pwm_start(self._channel, duty_cycle, frequency)

    def pwm_stop(self, channel):
        self._device.pwm_stop(self._channel)

    def value_changed(self, new_value, old_value):
        self.set(new_value)

    def kervi_value_changed(self, changed_input, value):
        if changed_input == self.pwm["duty_cycle"]:
            if self.pwm["active"].value:
                self._device.pwm_start(self._channel, duty_cycle=changed_input.value)

        if changed_input == self.pwm["frequency"]:
            self._device.pwm_start(self._channel, frequency=changed_input.value)

        if changed_input == self.pwm["active"]:
            if changed_input.value:
                self._device.pwm_start(self._channel)
            else:
                self._device.pwm_stop(self._channel)

class AnalogIOChannel(NumberValue):
    def __init__(self, gpio_device, channel, is_input):
        NumberValue.__init__(self, str(channel), is_input=is_input)
        self._device = gpio_device
        self._channel = channel
        self.min = 0
        self.max = 100

    def get(self):
        return self._device.get(self, self._channel)

    def set(self, value):
        self._device.set(self._channel, value)

    def define_as_input(self, pullup=False):
        self.is_input = False
        self._device.define_as_input(self._channel, pullup)

    def define_as_output(self):
        self.is_input = True
        self._device.define_as_output(self._channel)

    def value_changed(self, new_value, old_value):
        self.set(new_value)

class IGPIODeviceDriver(object):
    """
    """
    def __init__(self, gpio_id):
        self.channels = {}
        self._gpio_id = gpio_id

    def __getitem__(self, channel):
        channel = str(self._map_pin(channel))
        if not channel in self.channels.keys():
            channel_type = self._get_channel_type(channel)
            if channel_type == CHANNEL_TYPE_ANALOG_IN:
                self.channels[channel] = AnalogIOChannel(self, channel, False)
            elif channel_type == CHANNEL_TYPE_ANALOG_OUT:
                self.channels[channel] = AnalogIOChannel(self, channel, True)
            elif channel_type == CHANNEL_TYPE_GPIO:
                self.channels[channel] = LogicIOChannel(self, channel)
            else:
                KerviLog("gpio").warning("unknown channel_type: %s %s", channel, channel_type)

        return self.channels[channel]

    def _map_pin(self, channel):
        return channel

    def _get_channel_type(self, channel):
        """creates a channel"""
        raise NotImplementedError

    def define_as_input(self, channel, pullup=False):
        """Define a channel as input"""
        raise NotImplementedError

    def define_as_output(self, channel):
        """Define a channel as output"""
        raise NotImplementedError

    def define_as_pwm(self, channel, frequency, duty_cycle=None):
        """
        Defines a channel as pwm output.

        :param channel:
            The channel to define as a pwm output.

        :type channel: ``int``

        :param frequency:
            The pwn frequency.

        :type channel: ``int``

        :param duty_cycle:
            The duty_cycle to use, can be changed in call to pwm_start

        :type channel: ``int``
        """
        raise NotImplementedError

    def set(self, channel, state):
        """Sets the state of a channel that is defined as output.

        :param channel:
            The channel that should be changed.

        :type channel: ``int``

        :param state:
            The state of the channel.

        :type channel: ``bool``

        """
        raise NotImplementedError

    def set_channels(self, channels):
        """Sets the state of multiple channels in one operation.

        :param channels:
            A dictionary where keys are channels and values the value to set for each channel.

        :type channels: ``dict``

        """
        for key in channels:
            self.set(key, channels[key])

    def get(self, channel):
        """
        Returns the state of a channel.
        """
        raise NotImplementedError

    def listen(self, channel, callback, bounce_time=.2):
        """
        Listen on a channel for state change.
        The callback function is called when a channel is going high and low.

        :param channel:
            The channel to listen on.

        :type channel: ``int``

        :param callback:
            The function or method to call when the state of the channel change.

            .. code:: python
                def callback_func(state):
                    print(state)

                GPIO.listen(12, callback_func)

        :type callback: ``function or method``

        :param bounce_time:
            To reduce switch noice a bounce_time > 0 filters small jitters when a switch or button is pressed

        :type channel: ``float``
        """

        raise NotImplementedError

    def listen_rising(self, channel, callback):
        """
        calls the callback when a channel is going high.

        :param channel:
            The channel to listen on.

        :type channel: ``int``

        :param callback:
            The function or method to call when the state of the channel change.

        :type channel: ``function or method``
        """
        raise NotImplementedError

    def listen_falling(self, channel, callback):
        """
        calls the callback when a channel is going low.

        :param channel:
            The channel to listen on.

        :type channel: ``int``

        :param callback:
            The function or method to call when the state of the channel change.

        :type channel: ``function or method``
        """

        raise NotImplementedError

    def pwm_start(self, channel, duty_cycle=None, frequency=None):
        """
        Starts the pwm signal on a channel. The channel should be defined as pwm prior to this call.
        If no duty_cycle or frequency is passed in this call previous values from call to
        define_as_pwm or pwm_start is used.

        :param channel:
            The channel to start the pwm signal on.

        :type channel: ``int``

        :param duty_cycle:
            The duty cycle use on the channel.

        :type duty_cycle: ``int``

        :param frequency:
            The frequency to be used on the pwm channel.

        :type frequency: ``int``

        """
        raise NotImplementedError

    def pwm_stop(self, channel):
        """
        Stop pwm signal on channel.
        """
        raise NotImplementedError



#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

try:
    from datetime import datetime
except:
    from kervi.core.utility.udatetime import datetime

from kervi.core.utility.component import KerviComponent

from kervi.actions import Actions
VALUE_COUNTER = 0
class KerviValue(KerviComponent):
    """
    Generic value that is use as base class for specialized  kervi values.
    """
    def __init__(self, name, value_type, **kwargs):
        global VALUE_COUNTER
        parent = kwargs.get("parent", None)
        input_id = kwargs.get("value_id", None)
        if input_id is None:
            VALUE_COUNTER += 1
            if parent:
                input_id = parent.component_id + ".VALUE_" + str(VALUE_COUNTER)
            else:
                input_id = "VALUE_" + str(VALUE_COUNTER)
        elif parent:
            input_id = parent.component_id + "." + input_id

        KerviComponent.__init__(self, input_id, value_type, name, **kwargs)
        #self.spine = Spine()
        self._default_value = None
        self._index = kwargs.get("index", None)
        self.is_input = kwargs.get("is_input", True)
        self._value = None
        self._display_value = None
        self._unit = ""
        self._display_unit = None
        
        self._sparkline = []
        self._observers = []
        self._value_event_handlers = []
        if parent:
            self._observers += [parent]
        self._spine_observers = {}
        self.command = self.component_id + ".setValue"
        self.spine.register_command_handler(self.command, self._set_value, groups=self.admin_groups)
        self.spine.register_query_handler("getKerviValue", self._query_value, groups=self.user_groups)

        self._ui_parameters = {
            "size": 0,
            "input_size": 0,
            "type": "unkonwn",
            "link_to_header": False,
            "label_icon": None,
            "label": self.name,
            "label_size":0,
            "flat": False,
            "inline": False,
            "is_input":self.is_input,
            "value_size":3,
            "width":0,
            "height":0

        }
        self._persist_value = kwargs.get("persist_value", False)
        self._log_values = kwargs.get("log_values", False)

    @property
    def index(self):
        """
        Index of the value relative to its parent (controller or sensor)
        """
        return self._index

    @property
    def value_id(self):
        return self.component_id

    @property
    def value(self):
        """Current value of the component"""
        return self._value

    @value.setter
    def value(self, new_value):
        """
        Updates the value.
        If the change exceeds the change delta observers and linked values are notified.  
        """
        self._set_value(new_value)

    @property
    def display_value(self):
        return self._value

    
    @property
    def unit(self):
        """
        Metric Unit of value.

        :type: ``str``
        """
        return self._unit

    @unit.setter
    def unit(self, value):
        self._unit = value
    
    @property
    def display_unit(self):
        if self._display_unit:
            return self._display_unit
        else:
            return self._unit
    
    @property
    def log_values(self):
        """
        Set to true if the values should be logged to DB.
        If false Kervi will hold a small cache in memory
        for the last reading to be used in sparklines and real time charts.
        """
        return self._log_values

    @log_values.setter
    def log_values(self, value):
        self._log_values = value

    @property
    def persist_value(self):
        """
        Set to true if the current value should be saved to db when the Kervi application or module exits.
        The value will be restored upon application or module start. 
        """
        return self._persist_value

    @persist_value.setter
    def persist_value(self, do_persist):
        if not self._persist_value and do_persist:
            self._persist_value = do_persist
            self._load_persisted()
        else:
            self._persist_value = do_persist

    def _query_value(self, id):
        if self.component_id == id:
            return self.value

    def _load_persisted(self):
        if self._persist_value:
            value = self.settings.retrieve_value("value", self._default_value)
            self._set_value(value, False)

    def link_to(self, source, transformation=None):
        """
        Kervi values may be linked together. 
        A KerviValue is configured to be either an input or output.
        When an output value is linked to an input value the input will become
        an observer of the output. Every time the output value change it will notify the input 
        about the change.

        :param source:
            It is possible to make direct and indirect link.

            If the type of the source parameter is of type KerviValue a direct link is created.
            this is a fast link where the output can notify directly to input.
            This type of link is possible if both output and input resides in the same process.

            If the type of the source parameter is a string it is expected to hold the id of a another KerviValue.
            In this mode the input value will listen on the kervi spine for events from the output value.
            This mode is useful if the output and input does not exists in the same process.
            This type of link must be made by the input.

        :type source: ``str`` or ``KerviValue``

        :param transformation:
            A function or lambda expression that transform the value before the output is update.
            This is usefull if the input expects other ranges that the output produce
            or need to change the sign of the value.

        """
        if isinstance(source, KerviValue):
            if source.is_input and not self.is_input:
                self.add_observer(source, transformation)
            elif not source.is_input and self.is_input:
                source.add_observer(self, transformation)
            else:
                raise Exception("input/output mismatch in kervi value link:{0} - {1}".format(source.name, self.name))
        elif isinstance(source, str):
            if len(self._spine_observers) == 0:
                self.spine.register_event_handler("valueChanged", self._link_changed_event)

            self._spine_observers[source] = transformation

    def link_to_dashboard(self, dashboard_id=None, panel_id=None, **kwargs):
        r"""
        Links this value to a dashboard panel.

        :param dashboard_id:
            Id of the dashboard to link to.
        :type dashboard_id: str

        :param panel_id:
            Id of the panel on the dashboard to link to.
        :type panel_id: str
        
        :param \**kwargs:
            Use the kwargs below to override default values for ui parameters

        :Keyword Arguments:
            
            * *link_to_header* (``str``) -- Link this value to header of the panel.

            * *label_icon* (``str``) -- Icon that should be displayed together with label.

            * *label* (``str``) -- Label text, default value is the name of the value.

            * *flat* (``bool``) -- Flat look and feel.

            * *inline* (``bool``) -- Display value and label in its actual size
                If you set inline to true the size parameter is ignored.
                The value  will only occupy as much space as the label and input takes.

            * *input_size* (``int``) -- width of the slider as a percentage of the total container it sits in.

            * *value_size* (``int``) -- width of the value area as a percentage of the total container it sits in.
        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            panel_id,
            **kwargs
        )

    def add_observer(self, observer, transformation=None):
        """
        Adds an observer to this value.
        The observer must implement the function value_changed. 
        """
        self._observers += [(observer, transformation)]

    def _link_changed_event(self, id, source, old_value):
        if source["id"] in self._spine_observers.keys():
            transformation = self._spine_observers[source["id"]]
            if transformation:
                self.value = transformation(source["value"])
            else:
                self.value = source["value"]

    def _set_value(self, nvalue, allow_persist=True):
        if self._value != nvalue:
            self.spine.log.debug(
                "value change on input:{0} value:{1}",
                self.component_id,
                nvalue
            )

            old_value = self.value
            self._value = nvalue
            try:
                self.value_changed(nvalue, old_value)
            except Exception as Ex:
                self.spine.log.exception("Error in value changed")
            
            for observer in self._observers:
                try:
                    item, transformation = observer
                    if transformation:
                        item.kervi_value_changed(self, transformation(nvalue))
                    else:
                        item.kervi_value_changed(self, nvalue)
                except Exception:
                    self.spine.log.exception("error in value observer:")

            self._check_value_events(nvalue, old_value)

            if self._persist_value and allow_persist:
                self.settings.store_value("value", self.value)

            val = {
                "id":self.component_id,
                "value":nvalue,
                "timestamp":datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                "display_value": self.display_value,
                "display_unit": self.display_unit
            }
            self.spine.trigger_event(
                "valueChanged",
                self.component_id,
                val,
                self._log_values,
                groups=self.user_groups
            )

    def kervi_value_changed(self, source, value):
        self._set_value(value, False)

    def value_changed(self, new_value, old_value):
        pass

    def add_value_event(self, event_value, func, event_type=None, parameters=None, **kwargs):
        """
        Add a function that is called when the value reach or pass the event_value.

        :param event_value:
            A single value or range specified as a tuple.

            If it is a range the function specified in func is called when the value enters the range.

        :type event_value: ``float``, ``string``, ``boolean`` or a tuple of these types.

        :param func:
            Function or lambda expression to be called.
            This function will receive the dynamcic value as a parameter.

        :param event_type:
            String with the value "warning" of "error" or None (default).
            If warning or error is specified the value or range are shown in UI.

        :type event_type: ``str``

        """
        self._value_event_handlers += [(event_value, func, event_type, parameters, kwargs)]


    def _handle_range_event(self, value, message, func, level, **kwargs):
        message_type = ""
        message_color = "#28a745"
        if level==1:
            message_type = "Error"
            message_color = "#dc3545"
        elif level==2:
            message_type = "Warning"
            message_color = "#ffc107"
        if message:
            from kervi.messaging import Messaging
            from kervi.core.utility.superformatter import SuperFormatter
            sf = SuperFormatter()
            
            body_template = self.global_config.texts.messages.value_plain
            body = sf.format(
                body_template,
                message_color=message_color,
                message_type = message_type,
                source_name=self._name,
                message=message,
                value=self.display_value,
                sparkline=self._sparkline,
                user_name="{user_name}",
                unit=self._display_unit,
                level=level
            ) 

            html_template = self.global_config.texts.messages.value_html
            html_body = sf.format(
                html_template,
                message_color=message_color,
                message_type = message_type,
                source_name=self._name,
                message=message,
                value=self.display_value,
                sparkline=self._sparkline,
                user_name="{user_name}",
                level=level,
                unit=self._display_unit
            ) 

            kwargs = dict(kwargs, source_id=self.component_id, source_name=self.name, user_groups = self._user_groups, level=level, body=body, body_html=html_body)
            Messaging.send_message(message, **kwargs)
            
        if func:
            func(self)

    def add_normal_range(self, value, message=None, func=None, **kwargs):
        self.add_value_event(value, self._handle_range_event, parameters=[message, func, 3], **kwargs)

    def add_warning_range(self, value, message=None, func=None, **kwargs):
        self.add_value_event(value, self._handle_range_event, event_type="warning", parameters=[message, func, 2], **kwargs)

    def add_error_range(self, value, message=None, func=None, **kwargs):
        self.add_value_event(value, self._handle_range_event, event_type="error", parameters=[message, func, 1], **kwargs)

    def _check_value_events(self, new_value, old_value):
        for event in self._value_event_handlers:
            value, func, event_type, parameters, kwargs = event
            if func:
                if isinstance(value, tuple):
                    value_start, value_end = value
                    if old_value < value_start and new_value >= value_start:
                        func(self, *parameters, **kwargs)
                    elif old_value > value_end and new_value <= value_end:
                        func(self, *parameters, **kwargs)
                else:
                    if old_value < value and new_value >= value:
                        func(self, *parameters, **kwargs)
                    elif old_value > value and new_value <= value:
                        func(self, *parameters, **kwargs)

    @property
    def _event_ranges(self):
        ranges = []
        for event in self._value_event_handlers:
            value, func, event_type, parameters, kwargs = event
            if isinstance(value, tuple):
                value_start, value_end = value
                ranges += [{"start":value_start, "end":value_end, "type":event_type}]
            elif event_type:
                ranges += [{"start":value, "end":None, "type":event_type}]
        return ranges

    
    def _get_ui_parameters(self, ui_parameters):
        ui_parameters["display_unit"] = self.display_unit
        return ui_parameters
    
    def _get_info(self, **kwargs):
        return {
            "value":self.value,
            "command":self.command,
            "ranges":self._event_ranges,
            "visible": self.visible,
        }

    def on_get_value(self):
        return self.value

    def _get_bus_routes(self):
        result = [] 
        for link in self._spine_observers.keys():
            result += [{
                "id": link,
                "direction": "in",
                "topic_type": "event",
                "topic": "valueChanged"
                }]
        return result


#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

import random
try:
    from datetime import datetime
except:
    from kervi.core.utility.udatetime import datetime
import kervi.spine as spine
from kervi.core.utility.settings import Settings
from kervi.config import Configuration, Text

class DasboardSectionLink(object):
    def __init__(self, dashboard_id, section_id, section_caption, param, component):
        self.link_id = random.getrandbits(64)
        self.dashboard_id = dashboard_id
        self.section_id = section_id
        self.parameters = param
        self.component = component
        self.section_caption = section_caption

    def set_parameter(self, name, value):
        self.parameters[name] = value
        self.component.spine.triggerEvent("dashboardLinkChanged", self.link_id, self.parameters)

    

class KerviComponent(object):
    def __init__(self, component_id, component_type, name, **kwargs):
        self.spine = kwargs.get("spine", spine.Spine())
        self._configuration = kwargs.get("configuration", Configuration)
        self._component_id = component_id
        self._component_type = component_type
        self._admin_groups = kwargs.get("admin_groups", [])
        self._user_groups = kwargs.get("user_groups", [])
        self._user_groups += self.admin_groups
        self._name = name
        self._icon = None
        self._visible = True
        self._dashboard_links = []
        self._ui_parameters = {}
        self._settings = Settings(self.component_type + "_" + self.component_id)
        if self.spine:
            self.spine.log.debug("component created:{0}", self.component_id)
            self.spine.register_query_handler(
                "getDashboardComponents",
                self._get_dashboard_components,
                groups=self.user_groups
            )
            self.spine.register_query_handler("getComponentInfo", self._get_component_info, groups=self.user_groups)
            self.spine.register_query_handler("getComponentRoutes", self._get_component_bus_routes)

    @property
    def component_id(self):
        return self._component_id

    @property
    def component_type(self):
        return self._component_type

    @property
    def global_config(self):
        return self._configuration

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.spine.trigger_event("componentChangeName", self.component_id, value)

    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, value):
        self._icon = value
        self.spine.trigger_event("componentChangeIcon", self.component_id, value)

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, value):
        self._visible = value

    @property
    def user_groups(self):
        return self._user_groups

    @user_groups.setter
    def user_groups(self, value):
        self._user_groups.clear()
        self._user_groups += value
        self._user_groups += self.admin_groups

    @property
    def admin_groups(self):
        return self._admin_groups

    @admin_groups.setter
    def admin_groups(self, value):
        self._admin_groups.clear()
        self._admin_groups += value
        self._user_groups += self.admin_groups

    @property
    def ui_groups(self):
        resulting_list = list(self.user_groups)
        resulting_list.extend(x for x in self.admin_groups if x not in resulting_list)
        return resulting_list

    @property
    def settings(self):
        """
        You can persist settings to the kervi database via the settings property.

        .. code:: python
            myComponent.settings.store_value("nameOfValue", value)

        And retrieve it again via.

        .. code:: python
            value = myComponent.settings.retrieve_value("nameOfValue)


        """
        return self._settings

    def set_ui_parameter(self, name, value):
        if name in self._ui_parameters:
            self._ui_parameters[name] = value
        else:
            raise ValueError("invalid ui parameter name:" + name)

    def link_to_dashboard(self, dashboard_id=None, section_id=None, **kwargs):
        section_caption = None
        if not dashboard_id:
            dashboard_id="**default**"

        section_id = kwargs.get("panel_id", section_id)
        
        if not section_id:
            section_id = self.component_id
            section_caption =  self.name

        ui_param = {}
        for key, value in self._ui_parameters.items():
            ui_param[key] = value

        for key, value in kwargs.items():
            if key in ui_param:
                ui_param[key] = value
            else:
                raise ValueError("illigal ui parameter name:" + key)

        if "link_to_header" not in ui_param:
            ui_param["link_to_header"] = False

        if "icon" not in ui_param:
            ui_param["icon"] = None
        link = DasboardSectionLink(dashboard_id, section_id, section_caption, ui_param, self)
        self._dashboard_links += [link]
        return link

    def get_reference(self):
        """Returns a reference for this component"""
        return {
            "id": self.component_id,
            "component_type": self.component_type
        }

    def _get_info(self, **kwargs):
        """Abstract method that must return properties for the component as a dictionary
        .. document private functions
        """
        self.spine.log.debug(
            "abstract get_info reached:{0}",
            self.component_id,
        )
        return {}

    def _get_ui_parameters(self, ui_parameters):
        return ui_parameters

    def _get_component_info(self, component_id=None, **kwargs):
        if  component_id is None or component_id == self.component_id:
            session = kwargs.get("session", None)
            authorized = True
            if session and len(self.user_groups) > 0:
                for group in self.user_groups:
                    if group in session["groups"]:
                        break
                else:
                    authorized = False
                
            if authorized:
                info = self._get_info(**kwargs)
                ui_parameters = self._get_ui_parameters(self._ui_parameters)
                info["componentType"] = self.component_type
                info["id"] = self.component_id
                info["visible"] = self.visible
                info["name"] = self.name
                info["ui"] = self._camel_case_parameters(ui_parameters)
                dashboard_links = []
                for link in self._dashboard_links:
                    param = self._camel_case_parameters(link.parameters)
                    dashboard_links += [{
                        "linkId": link.link_id,
                        "dashboardId": link.dashboard_id,
                        "sectionId": link.section_id,
                        "sectionName": link.section_caption,
                        "parameters":param
                    }]

                info["dashboardLinks"] = dashboard_links
                return info


    def _underscore_to_camelcase(self, value):
        def _camelcase():
            yield str.lower
            while True:
                yield str.capitalize

        camelcase = _camelcase()
        return "".join(next(camelcase)(x) if x else '_' for x in value.split("_"))

    def _camel_case_parameters(self, parameters):
        result = {}
        for key in parameters:
            result[self._underscore_to_camelcase(key)] = parameters[key]
        return result

    def _get_dashboard_components(self, dashboard_id, section_id):
        result = []
        for link in self._dashboard_links:
            if ((link.dashboard_id == "*" or link.dashboard_id == dashboard_id)
                    and link.section_id == section_id):
                param = self._camel_case_parameters(link.parameters)
                result += [{
                    "linkId": link.link_id,
                    "componentId":self.component_id,
                    "parameters":param
                }]
        
        return result

    def _get_component_bus_routes(self):
        result = self._get_bus_routes()
        return {
            "id": self._component_id,
            "routes": result,
        }

    def _get_bus_routes(self):
        result = [] 
        for link in self._dashboard_links:
            result += [{
                "id": link.link_id,
                "direction": "out",
                "topic_type": "event",
                "topic": "valueChanged"
                }]
        return result


#Copyright 2018 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

from kervi.actions.action import _LinkedAction

class Actions(object):
    
    __actions = {}
    __unbound_actions = {}
    __unbound_interrupts = {}
    def __init__(self):
        pass

    def add(self, action):
        Actions.__actions[action.action_id] = action

    def add_unbound(self, handler_name, action_id, name):
        Actions.__unbound_actions[handler_name] = (action_id, name)

    def add_unbound_interrupt(self, interrupt_name, action_id):
        Actions.__unbound_interrupts[interrupt_name] = (action_id)

    def is_unbound(self, handler_name):
        return handler_name in Actions.__unbound_actions

    def is_unbound_interrupt(self, interrupt_name):
        return interrupt_name in Actions.__unbound_interrupts

    def get_unbound(self, handler_name):
        return Actions.__unbound_actions[handler_name]

    def get_unbound_interrupt(self, interrupt_name):
        return Actions.__unbound_interrupts[interrupt_name]

    def __getitem__(self, action_id):
        if action_id in Actions.__actions:
            return Actions.__actions[action_id]
        else:
            linked_action = _LinkedAction(action_id)
            Actions.__actions[action_id] = linked_action
            return linked_action

